/* PS/2 mouse to Atari ST adapter. (C) 2006 - 2016 Christian Zietz <czietz@gmx.net> */
// -- Main code

/* PLEASE NOTE: Modified in 2016 to compile with recent avr-gcc/avr-libc. 
   Modifications untested! Use at your own risk! 
*/

/*

Copyright (c) 2016-2016 Christian Zietz <czietz@gmx.net> 

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/

#include <stdlib.h>
#include <string.h>
#include <avr/io.h>

#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "ps2.h"
#include "debug.h"

#define REVISION "C"

// -- output defines

#define sgn(x) ((x>0)?1:-1);

#define lbut 0x08
#define rbut 0x20

#define xmask 0x03
#define ymask 0x14

const unsigned char grayx[4] PROGMEM = {0x00,0x01,0x03,0x02};
const unsigned char grayy[4] PROGMEM = {0x00,0x04,0x14,0x10};

#define SPD_MAX 252
#define SPD_LEN 30
const unsigned char spds[SPD_LEN] PROGMEM = {0,104,180,205,218,226,231,234,237,239,241,242,243,244,245,246,247,247,248,248,248,249,249,249,250,250,250,250,251,251};

// #define SPD_MAX 253
// #define SPD_LEN 19
//unsigned char spds[SPD_LEN] PROGMEM = {0x00, 0x60, 0x70, 0x80, 0x90, 0xa0, 0xb0, 0xc0, 0xd0, 0xd8, 0xe0, 0xe8, 0xf0, 0xf4, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc};

//#define SPD_MAX 253
//#define SPD_LEN 30
//unsigned char spds[SPD_LEN] PROGMEM = {0, 128, 192, 213, 224, 230, 235, 238, 240, 242, 243, 244, 245, 246, 247, 247, 248, 248, 249, 249, 250, 250, 250, 250, 251, 251, 251, 251, 252, 252};

unsigned char xspd, yspd;
unsigned char xcnt, ycnt, xt;
unsigned char xdir, ydir;
//--


SIGNAL(TIMER0_OVF_vect) {
  unsigned char oldc;
  
  xcnt += xdir;
  xcnt &= 3;
  
  oldc = PORTC;
  oldc &= ~xmask;
  oldc |= pgm_read_byte(&(grayx[xcnt]));
  PORTC = oldc;
  
  TCNT0 = xspd;
}

SIGNAL(TIMER2_OVF_vect) {
  unsigned char oldc;
  
  ycnt += ydir;
  ycnt &= 3;
  
  oldc = PORTC;
  oldc &= ~ymask;
  oldc |= pgm_read_byte(&(grayy[ycnt]));
  PORTC = oldc;
  
  TCNT2 = yspd;
}

SIGNAL(TIMER1_COMPA_vect) {
#ifdef DEBUG
  static int i;
  char b[5];
#endif
  int st, x, y;
  
  // this may take some time, enable interrupts, so we don't loose them
  sei();
  ps2send(0xeb);
  ps2recv();
  st = ps2recv();
  x = ps2recv();
  y = ps2recv(); 
  cli();
  
  x -= (0x10*(st&0x10));
  y -= (0x08*(st&0x20));
  
#ifdef DEBUG
  i++;
  if (i&1) {
    debug_send_str("\nST:");
    itoa(st, b, 16);
    debug_send_str(b);
    debug_send_str("X:");
    itoa(x, b, 10);
    debug_send_str(b);
    debug_send_str("Y:");
    itoa(y, b, 10);
    debug_send_str(b); 
  }
#endif   
  
  // note that we now switch only DDR for mouse buttons to switch between high impedance and low to mimic the original mouse
  if (st & 0x01) {
    DDRC |= lbut;
  } else {
    DDRC &= ~lbut;
  }
  
  if (st & 0x02) {
    DDRC |= rbut;
  } else {
    DDRC &= ~rbut;
  }
  
  if (x!=0) {
    if (abs(x)>=SPD_LEN) {
      xspd = SPD_MAX;
    } else {
      xspd = pgm_read_byte(&spds[abs(x)]);
    }
    xdir = -sgn(x);
    TCNT0 = xspd;
  } else {
    xdir = 0;
  }

  if (y!=0) {
    if (abs(y)>=SPD_LEN) {
      yspd = SPD_MAX;
    } else {
      yspd = pgm_read_byte(&spds[abs(y)]);
    }
    ydir = -sgn(y);
    TCNT2 = yspd;
  } else {
    ydir = 0;
  }
}

int main(void) {
  PORTC = 0xff - lbut - rbut;
  DDRC = 0x3f - lbut - rbut; // put mouse button outputs into high impedance state
  
  PORTD = 0xff;

#ifdef DEBUG  
  enable_debug();
  debug_send_str("  \nCHZ PS2Atari Rev. " REVISION);
#endif  
  
  // timer 0
  TCCR0  |= (1<<CS02)+(1<<CS00);
  TIMSK  |= (1<<TOIE0);
 
  // timer 1
  OCR1A   = 150;
  TCCR1B |= (1<<WGM12);
  TCCR1B |= (1<<CS12) + (1<<CS10);
  TIMSK  |= (1<<OCIE1A);
 
  // timer 2
  TCCR2  |= (1<<CS22)+(1<<CS21)+(1<<CS20);
  TIMSK  |= (1<<TOIE2);
  
  // global variables
  xdir = ydir = xcnt = ycnt = 0;
  
  ps2send(0xff);
  ps2recv();
  ps2recv();
  ps2recv();
  ps2send(0xf0);
  ps2recv();
  sei();
  while (1) {}
}
